/*
 * http.c : GeeXboX uShare Web Server handler.
 * Originally developped for the GeeXboX project.
 * Parts of the code are originated from GMediaServer from Oskar Liljeblad.
 * Copyright (C) 2005-2007 Benjamin Zores <ben@geexbox.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <pthread.h> 
#include <signal.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <upnp/upnp.h>
#include <upnp/upnptools.h>

#include "services.h"
#include "cds.h"
#include "cms.h"
#include "msr.h"
#include "metadata.h"
#include "http.h"
#include "minmax.h"
#include "trace.h"
#include "presentation.h"
#include "osdep.h"
#include "mime.h"
#include "fifo.h"

#define PROTOCOL_TYPE_PRE_SZ  11   /* for the str length of "http-get:*:" */
#define PROTOCOL_TYPE_SUFF_SZ 2    /* for the str length of ":*" */

struct web_file_t {
  char *fullpath;
  off_t pos;
  enum {
    FILE_LOCAL,
    FILE_MEMORY,
    FILE_STREAM
  } type;
  union {
    struct {
      int fd;
      struct upnp_entry_t *entry;
    } local;
    struct {
      char *contents;
      off_t len;
    } memory;
  } detail;
};


static inline void
set_info_file (struct File_Info *info, const size_t length,
               const char *content_type)
{
  info->file_length = length;
  info->last_modified = 0;
  info->is_directory = 0;
  info->is_readable = 1;
  info->content_type = ixmlCloneDOMString (content_type);
}

static int
http_get_info (const char *filename, struct File_Info *info)
{
  extern struct ushare_t *ut;
  struct upnp_entry_t *entry = NULL;
  struct stat st;
  int upnp_id = 0;
  char *content_type = NULL;
  char *protocol = NULL;
  
  if (!filename || !info)
    return -1;

  log_verbose ("http_get_info, filename : %s\n", filename);

  if (!strcmp (filename, CDS_LOCATION))
  {
    set_info_file (info, CDS_DESCRIPTION_LEN, SERVICE_CONTENT_TYPE);
    return 0;
  }

  if (!strcmp (filename, CMS_LOCATION))
  {
    set_info_file (info, CMS_DESCRIPTION_LEN, SERVICE_CONTENT_TYPE);
    return 0;
  }

  if (!strcmp (filename, MSR_LOCATION))
  {
    set_info_file (info, MSR_DESCRIPTION_LEN, SERVICE_CONTENT_TYPE);
    return 0;
  }

  if (ut->use_presentation && !strcmp (filename, USHARE_PRESENTATION_PAGE))
  {
    if (build_presentation_page (ut) < 0)
      return -1;

    set_info_file (info, ut->presentation->len, PRESENTATION_PAGE_CONTENT_TYPE);
    return 0;
  }

  if (ut->use_presentation && !strncmp (filename, USHARE_CGI, strlen (USHARE_CGI)))
  {
    if (process_cgi (ut, (char *) (filename + strlen (USHARE_CGI) + 1)) < 0)
      return -1;

   set_info_file (info, ut->presentation->len, PRESENTATION_PAGE_CONTENT_TYPE);
    return 0;
  }

  upnp_id = atoi (strrchr (filename, '/') + 1);
  entry = upnp_get_entry (ut, upnp_id);
  if (!entry)
    return -1;

  if (!entry->fullpath)
    return -1;

  if (stat (entry->fullpath, &st) < 0)
    return -1;

  if (access (entry->fullpath, R_OK) < 0)
  {
    if (errno != EACCES)
      return -1;
    info->is_readable = 0;
  }
  else
    info->is_readable = 1;

  /* file exist and can be read */
  info->file_length = st.st_size;
  info->last_modified = st.st_mtime;
  info->is_directory = S_ISDIR (st.st_mode);

  //Override the size entry for m3u files - we will supply a stream, not data from the file
  if(!strcmp(strrchr(filename, '\0') - 4, ".m3u")){
    info->file_length = (off_t)FIFO_STREAM_LENGTH;
  }
    
  protocol = 
#ifdef HAVE_DLNA
    entry->dlna_profile ?
    dlna_write_protocol_info (DLNA_PROTOCOL_INFO_TYPE_HTTP,
                              DLNA_ORG_PLAY_SPEED_NORMAL,
                              DLNA_ORG_CONVERSION_NONE,
                              DLNA_ORG_OPERATION_RANGE,
                              ut->dlna_flags, entry->dlna_profile) :
#endif /* HAVE_DLNA */
    mime_get_protocol (entry->mime_type);

  content_type =
    strndup ((protocol + PROTOCOL_TYPE_PRE_SZ),
             strlen (protocol + PROTOCOL_TYPE_PRE_SZ)
             - PROTOCOL_TYPE_SUFF_SZ);
  free (protocol);

  if (content_type)
  {
    info->content_type = ixmlCloneDOMString (content_type);
    free (content_type);
  }
  else
    info->content_type = ixmlCloneDOMString ("");

  return 0;
}

static UpnpWebFileHandle
get_file_memory (const char *fullpath, const char *description,
                 const size_t length)
{
  struct web_file_t *file;

  file = malloc (sizeof (struct web_file_t));
  file->fullpath = strdup (fullpath);
  file->pos = 0;
  file->type = FILE_MEMORY;
  file->detail.memory.contents = strdup (description);
  file->detail.memory.len = length;

  return ((UpnpWebFileHandle) file);
}

static UpnpWebFileHandle
http_open (const char *filename, enum UpnpOpenFileMode mode)
{
  extern struct ushare_t *ut;
  struct upnp_entry_t *entry = NULL;
  struct web_file_t *file;
  int fd, upnp_id = 0;

  if (!filename)
    return NULL;

  log_verbose ("http_open, filename : %s\n", filename);

  if (mode != UPNP_READ)
    return NULL;

  if (!strcmp (filename, CDS_LOCATION))
    return get_file_memory (CDS_LOCATION, CDS_DESCRIPTION, CDS_DESCRIPTION_LEN);

  if (!strcmp (filename, CMS_LOCATION))
    return get_file_memory (CMS_LOCATION, CMS_DESCRIPTION, CMS_DESCRIPTION_LEN);

  if (!strcmp (filename, MSR_LOCATION))
    return get_file_memory (MSR_LOCATION, MSR_DESCRIPTION, MSR_DESCRIPTION_LEN);

  if (ut->use_presentation && ( !strcmp (filename, USHARE_PRESENTATION_PAGE)
      || !strncmp (filename, USHARE_CGI, strlen (USHARE_CGI))))
    return get_file_memory (USHARE_PRESENTATION_PAGE, ut->presentation->buf,
                            ut->presentation->len);

  upnp_id = atoi (strrchr (filename, '/') + 1);
  entry = upnp_get_entry (ut, upnp_id);
  if (!entry)
    return NULL;

  if (!entry->fullpath)
    return NULL;

  log_verbose ("Fullpath : %s\n", entry->fullpath);

  // For streams to be buffered and relayed from the net - not serving m3u files  
  if(!strcmp(strrchr(filename, '\0') - 4, ".m3u")){
    
    int i = 0;
    
    if(my_fifo->tid && my_fifo->upnp_id!=upnp_id){
      //Existing stream download thread is running but upnp_id has changed
      //indicating URL requested has changed - close that existing connection 
      my_fifo->terminate = 1;
      
      while(my_fifo->tid){
	sleep (1);
	if(i > 15){
	  //Graceful termination has failed, try this instead. 
	  pthread_cancel(my_fifo->tid);
	  my_fifo->tid = (pthread_t)NULL;
	}
	i++;
      }
    }
    
    if(my_fifo->tid==(pthread_t)NULL && my_fifo->startingup==0){
      // No streaming download exists - initialise fifo and create thread.
      
      my_fifo->startingup = 1; // Prevent multiple rapid requests before threadID known
      
      // Find the URL to start streaming
      FILE* fp=fopen(entry->fullpath,"r");
      if (!fp){
	my_fifo->startingup = 0;
	return NULL;
      }
      
      char buf[256];
      char url[FIFO_URL_SIZE];
      int line = 0;
      
      url[0] = '\0';
      
      while(fgets(buf,sizeof(buf),fp)){
	line++;
	if(line==entry->line && !strncmp(buf, "http://", 7)){
	  sprintf (url, "%.*s", FIFO_URL_SIZE-1, buf+7);
	  url[strcspn(url, "\r\n")] = '\0'; // Remote new lines & line feeds
	  break;
	}
      }
      
      fclose(fp);
      
      if(strlen(url)==0){
	my_fifo->startingup = 0;
	return NULL;
      }

      // Initialise fifo and create thread. 
      fifo_init(my_fifo, url, upnp_id); 
      pthread_create(&my_fifo->tid, NULL, ReadStream, (void*)my_fifo);
    }
    
    file = malloc (sizeof (struct web_file_t));
    file->fullpath = strdup (entry->fullpath);
    file->pos = 0;
    file->type = FILE_STREAM;
    file->detail.local.entry = entry;
    file->detail.local.fd = (size_t)NULL;
    
  }else{
    // For real files shared from the drive
    fd = open (entry->fullpath, O_RDONLY | O_NONBLOCK | O_SYNC | O_NDELAY);
    if (fd < 0)
      return NULL;
    
    file = malloc (sizeof (struct web_file_t));
    file->fullpath = strdup (entry->fullpath);
    file->pos = 0;
    file->type = FILE_LOCAL;
    file->detail.local.entry = entry;
    file->detail.local.fd = fd; 
  }
  
  return ((UpnpWebFileHandle) file);
}

static int
http_read (UpnpWebFileHandle fh, char *buf, size_t buflen)
{
  struct web_file_t *file = (struct web_file_t *) fh;
  ssize_t len = -1;

  log_verbose ("http_read\n");

  if (!file)
    return -1;

  switch (file->type)
  {
  case FILE_LOCAL:
    log_verbose ("Read local file.\n");
    len = read (file->detail.local.fd, buf, buflen);
    break;
  case FILE_MEMORY:
    log_verbose ("Read file from memory.\n");
    len = (size_t) MIN (buflen, file->detail.memory.len - file->pos);
    memcpy (buf, file->detail.memory.contents + file->pos, (size_t) len);
    break;
  case FILE_STREAM:
    log_verbose ("Reading internet stream(1).\n");
    
    if(my_fifo->initialise == 1 || my_fifo->initialise == 2){
      //First stage initialisation data - the radio wants to be sent 128k inc ID3 tag, but is happy with null data
      //Second stage initialiation data - radio wants the last 128 bytes of the 'file', but is happy with null data
      memset (buf, 0, buflen);
      len = buflen;
      my_fifo->initialise--;
    }else{
      log_verbose("Sending data. Remaining buffer: %i.\n", my_fifo->unread);
      len = AccessFIFO(my_fifo, buf, buflen);
    }
    
    log_verbose ("Reading internet stream(2).\n");
    
    break;
  default:
    log_verbose ("Unknown file type.\n");
    break;
  }

  if (len >= 0)
    file->pos += len;

  log_verbose ("Read %zd bytes.\n", len);

  return len;
}

static int
http_write (UpnpWebFileHandle fh __attribute__((unused)),
            char *buf __attribute__((unused)),
            size_t buflen __attribute__((unused)))
{
  log_verbose ("http write\n");

  return 0;
}

static int
http_seek (UpnpWebFileHandle fh, off_t offset, int origin)
{
  struct web_file_t *file = (struct web_file_t *) fh;
  off_t newpos = -1;

  log_verbose ("http_seek\n");
  
  if (!file)
    return -1;

  switch (origin)
  {
  case SEEK_SET:
    log_verbose ("Attempting to seek to %ld (was at %ld) in %s\n",
                offset, file->pos, file->fullpath);
    newpos = offset;
    break;
  case SEEK_CUR:
    log_verbose ("Attempting to seek by %ld from %ld in %s\n",
                offset, file->pos, file->fullpath);
    newpos = file->pos + offset;
    break;
  case SEEK_END:
    log_verbose ("Attempting to seek by %ld from end (was at %ld) in %s\n",
                offset, file->pos, file->fullpath);

    if (file->type == FILE_LOCAL)
    {
      struct stat sb;
      if (stat (file->fullpath, &sb) < 0)
      {
        log_verbose ("%s: cannot stat: %s\n",
                    file->fullpath, strerror (errno));
        return -1;
      }
      newpos = sb.st_size + offset;
    }
    else if (file->type == FILE_MEMORY)
      newpos = file->detail.memory.len + offset;
    else if (file->type == FILE_STREAM)
      newpos = FIFO_STREAM_LENGTH + offset;
    break;
  }

  switch (file->type)
  {
  case FILE_LOCAL:
    /* Just make sure we cannot seek before start of file. */
    if (newpos < 0)
    {
      log_verbose ("%s: cannot seek: %s\n", file->fullpath, strerror (EINVAL));
      return -1;
    }

    /* Don't seek with origin as specified above, as file may have
       changed in size since our last stat. */
    if (lseek (file->detail.local.fd, newpos, SEEK_SET) == -1)
    {
      log_verbose ("%s: cannot seek: %s\n", file->fullpath, strerror (errno));
      return -1;
    }
    break;
  case FILE_MEMORY:
    if (newpos < 0 || newpos > file->detail.memory.len)
    {
      log_verbose ("%s: cannot seek: %s\n", file->fullpath, strerror (EINVAL));
      return -1;
    }
    break;
  case FILE_STREAM:
    //Does nothing for a stream 
    break;
  }

  file->pos = newpos;

  return 0;
}

static int
http_close (UpnpWebFileHandle fh)
{
  struct web_file_t *file = (struct web_file_t *) fh;

  log_verbose ("http_close\n");

  if (!file)
    return -1;

  switch (file->type)
  {
  case FILE_LOCAL:
    close (file->detail.local.fd);
    break;
  case FILE_MEMORY:
    /* no close operation */
    if (file->detail.memory.contents)
      free (file->detail.memory.contents);
    break;
  case FILE_STREAM:
    break;
  default:
    log_verbose ("Unknown file type.\n");
    break;
  }

  if (file->fullpath)
    free (file->fullpath);
  free (file);

  return 0;
}

struct UpnpVirtualDirCallbacks virtual_dir_callbacks =
  {
    http_get_info,
    http_open,
    http_read,
    http_write,
    http_seek,
    http_close
  };

// Access the buffer to read it, pass data out for other uses
int AccessFIFO(fifo_t * f, char *buf, int buflen)
{
     int bytes = 0;
     int brk = 0;
     int balance = buflen*1.25;
     int send,i,j;

     char frame[] = { 0xff, 0xfb, 0x42, 0xc4, 0x26, 0x3, 0x49, 0x7d, 0x83, 0x6, 0x40, 0x84, 0x69, 0xc1, 0x1b, 0x90, 0xa1, 0x8, 0x10, 0x8d, 0x29, 0x7c, 0xf5, 0xf7, 0xbf, 0xfc, 0xdf, 0x5a, 0xa1, 0xe2, 0x8, 0xd, 0x8a, 0x48, 0xd, 0x81, 0x84, 0x71, 0x85, 0x48, 0x29, 0xf, 0xf3, 0x27, 0x64, 0x46, 0x1d, 0xfe, 0xff, 0xfb, 0x7c, 0xaf, 0xe1, 0x37, 0x3f, 0xb7, 0xff, 0xa6, 0x23, 0x48, 0x3b, 0x6c, 0x2a, 0xfe, 0x3b, 0xef, 0x57, 0xe4, 0x5, 0xb7, 0x6a, 0xa1, 0x2d, 0x89, 0x52, 0x1a, 0xba, 0x7b, 0x14, 0x10, 0xb2, 0x56, 0x6d, 0xbd, 0x69, 0xdc, 0xd1, 0xd1, 0x98, 0xd6, 0xa5, 0xa6, 0x0, 0x0, 0xcd, 0xc6, 0xd1, 0x0, 0x7, 0x17, 0xe7, 0xce, 0xa, 0xd7, 0x30, 0x24, 0x64, 0x5f, 0x9f, 0x4c, 0xf3, 0x2a, 0x6c, 0x57, 0x36, 0xae, 0x46, 0xa, 0xd7, 0x84, 0xa6, 0xb, 0x91, 0x8f, 0xae, 0xe5, 0x17, 0x2a, 0x31, 0xd6, 0x91, 0xd8, 0xd2, 0x80, 0xf, 0x7e, 0xff, 0xef, 0xf0, 0x75, 0xc5, 0xef, 0x55, 0x7e, 0x1d, 0x41, 0x1, 0xc9, 0xd4, 0xcb, 0x27, 0x8e, 0xe4, 0x49, 0xb, 0xba, 0x16, 0x87, 0xda, 0xc7, 0x8, 0xd3, 0x77, 0xd5, 0xe6, 0xb6, 0xed, 0xeb, 0x32, 0x23, 0xd6, 0x96, 0x53, 0xd9, 0x5d, 0x8e, 0xac, 0x65, 0xbc, 0xd7, 0xc9, 0x2b, 0xa0 };
     
     
     if(f->buffersent==0){
     /*   bytes = 418; */
     /*   memset (buf, 0, bytes); */
     /*   buf[0] = '\xFF'; */
     /*   buf[1] = '\xFB'; */
     /*   buf[2] = '\x92'; */
     /*   buf[3] = '\x04'; */
       sleep(8);
     }









     
     
     /* } */
     //printf("Accessing FIFO\n");   


     /* if(f->unread > 0){ */

     /*   send = 0.02*f->unread; */

     /*   if(buflen > f->unread){ */
     /* 	 //send = 0.02*f->unread; */
	 
     /* 	 /\* bytes = 418; *\/ */
     /* 	 /\* memset (buf, 0, bytes); *\/ */
     /* 	 /\* buf[0] = '\xFF'; *\/ */
     /* 	 /\* buf[1] = '\xFB'; *\/ */
     /* 	 /\* buf[2] = '\x92'; *\/ */
     /* 	 /\* buf[3] = '\x04'; *\/ */
     /* 	 //sleep(2); */
	 
     /* 	 memset (buf, 0, buflen);  */
     /* 	 i=0; */
     /* 	 bytes = 0; */
     /* 	 while(bytes < buflen-418){ */
     /* 	   j = 418*i; */
     /* 	   buf[j] = '\xFF'; */
     /* 	   buf[j+1] = '\xFB'; */
     /* 	   buf[j+2] = '\x92'; */
     /* 	   buf[j+3] = '\x04'; */
     /* 	   bytes += 418; */
     /* 	   i++; */
     /* 	 } */
     /* 	 bytes= buflen; */
     /* 	 //Need to make up the difference */
	 
	 

	 
     /*   }else{ */
     /* 	 send = buflen; */
     /* 	 bytes = fifo_read(f, buf, send); */
     /*   } */
       
     /*   //Send a proportion of what we've got, up to the maximum. */
     /*   //bytes = fifo_read(f, buf, send); */
     /*   printf("Sending: %i. Buffer: %i. Skips: %i\n", bytes, f->unread, f->skip_read); */

     /* }else{ */
     /*   bytes = 0; */
     /* } */


     if(f->unread > buflen){
	 bytes = fifo_read(f, buf, buflen);      
       //Send a proportion of what we've got, up to the maximum.
       //bytes = fifo_read(f, buf, send);
       printf("Sending real data: %i. Buffer: %i. Skips: %i\n", bytes, f->unread, f->skip_read);

     }else{

       /* FILE* fp = fopen("/home/dif/programs/Ushare/2s_silence_56k_cut.mp3", "r"); */
       /* bytes = fread(buf, 1, buflen, fp); */
       
       /* fclose(fp); */

       i = 0;
       j = 0;
       /* while(i < 2000){  */
       /* 	 printf("%hhx ", buf[i]); */
       /* 	 i++; */
       /* } */

       bytes = 0;
       while(i<(buflen-1)){
	 //strcpy(&buf[i], &frame[j]);

	 buf[i] = frame[j];

	 if(i < 400)
	   printf("%hhx ", buf[i]); 
	 
	 i++;
	 if(j<(sizeof(frame)-1)){
	   j++;
	 }else{
	   j=0;
	 }
	 bytes++;
       }

              /* while(i < 2000){  */
       /* 	 printf("%hhx ", buf[i]); */
       /* 	 i++; */
       /* } */

       
       /* memset (buf, '\x00', buflen); */
       /* i=35; */

       /* //56kB silent file */
       /* /\* while(i < buflen){ *\/ */
       /* /\* 	 buf[i-3] = '\xFF'; *\/ */
       /* /\* 	 buf[i-2] = '\xFB'; *\/ */
       /* /\* 	 buf[i-1] = '\x40'; *\/ */
       /* /\* 	 buf[i] = '\xC4'; *\/ */
       /* /\* 	 i += 1280;    *\/ */
       /* /\* } *\/ */
 
       /* /\* //320kB silent file *\/ */
       /* /\* while(i < buflen){ *\/ */
       /* /\* 	 buf[i-3] = '\xFF'; *\/ */
       /* /\* 	 buf[i-2] = '\xFB'; *\/ */
       /* /\* 	 buf[i-1] = '\xE0'; *\/ */
       /* /\* 	 buf[i] = '\xC4'; *\/ */
       /* /\* 	 printf("%i: %hhx %hhx %hhx %hhx %hhx\n", i-3, buf[i-3], buf[i-2], buf[i-1], buf[i], buf[i+1]); *\/ */
       /* /\* 	 i += 10449; *\/ */
       /* /\* } *\/ */


       /* //320kB silent file = https://music.stackexchange.com/questions/62809/how-to-create-silence-in-mp3 */
       /* while(i < buflen){ */
       /* 	 buf[i-35] = '\xFF'; */
       /* 	 buf[i-34] = '\xFB'; */
       /* 	 buf[i-33] = '\x90'; */
       /* 	 buf[i-32] = '\x64'; */
       /* 	 if(i==35){ */
       /* 	   buf[i-31] = '\x00'; */
       /* 	   buf[i-30] = '\x0F'; */
       /* 	 }else{ */
       /* 	   buf[i-31] = '\x40'; */
       /* 	   buf[i-30] = '\x8F'; */
       /* 	 } */
       /* 	 buf[i-29] = '\xF0'; */
       /* 	 buf[i-28] = '\x00'; */
       /* 	 buf[i-27] = '\x00'; */
       /* 	 buf[i-26] = '\x69'; */
       /* 	 buf[i-25] = '\x00'; */
       /* 	 buf[i-24] = '\x00'; */
       /* 	 buf[i-23] = '\x00'; */
       /* 	 buf[i-22] = '\x08'; */
       /* 	 buf[i-21] = '\x00'; */
       /* 	 buf[i-20] = '\x00'; */
       /* 	 buf[i-19] = '\x0D'; */
       /* 	 buf[i-18] = '\x20'; */
       /* 	 buf[i-17] = '\x00'; */
       /* 	 buf[i-16] = '\x00'; */
       /* 	 buf[i-15] = '\x01'; */
       /* 	 buf[i-14] = '\x00'; */
       /* 	 buf[i-13] = '\x00'; */
       /* 	 buf[i-12] = '\x01'; */
       /* 	 buf[i-11] = '\xA4'; */
       /* 	 buf[i-10] = '\x00'; */
       /* 	 buf[i-9] = '\x00'; */
       /* 	 buf[i-8] = '\x00'; */
       /* 	 buf[i-7] = '\x20'; */
       /* 	 buf[i-6] = '\x00'; */
       /* 	 buf[i-5] = '\x00'; */
       /* 	 buf[i-4] = '\x34'; */
       /* 	 buf[i-3] = '\x80'; */
       /* 	 buf[i-2] = '\x00'; */
       /* 	 buf[i-1] = '\x00'; */
       /* 	 buf[i] = '\x04';   */
       /* 	 printf("%i: %hhx %hhx %hhx %hhx %hhx\n", i-35, buf[i-35], buf[i-34], buf[i-33], buf[i-32], buf[i-31]); */
       /* 	 i += 418; */
       /* } */
      
       /* bytes= buflen; */
       
       printf("Sending silent data: %i. Buffer: %i. Skips: %i\n", bytes, f->unread, f->skip_read);
     }



     
     /* while(!brk && f->skip_read <= 30)
     /*   {
     /* 	 if(f->unread <= balance){ */
     /* 	   //Too little data in the FIFO */
     /* 	   f->skip_read++; */
     /* 	   sleep(4); */
     /* 	   //printf("Skipping writing data from the FIFO. Now holding %i\n", f->unread); */
	   
     /* 	   send = 0.02*f->unread; */
     /* 	   //Send a proportion of what we've got, up to the maximum. */
     /* 	   bytes = fifo_read(f, buf, (((buflen)<(send))?(buflen):(send))); */
     /* 	   printf("Sending: %i. Buffer: %i. Skips: %i\n", bytes, f->unread, f->skip_read); */
     /* 	   brk = 1; */
     /* 	   /\* bytes = 418; *\/ */
     /* 	   /\* memset (buf, 0, bytes); *\/ */
     /* 	   /\* buf[0] = '\xFF'; *\/ */
     /* 	   /\* buf[1] = '\xFB'; *\/ */
     /* 	   /\* buf[2] = '\x92'; *\/ */
     /* 	   /\* buf[3] = '\x04'; *\/ */
	   
     /* 	   /\* buf[418] = '\xFF'; *\/ */
     /* 	   /\* buf[419] = '\xFB'; *\/ */
     /* 	   /\* buf[420] = '\x92'; *\/ */
     /* 	   /\* buf[421] = '\x04'; *\/ */
     /* 	   //if(f->skip_read == 10 || f->skip_read == 20){ */
     /* 	   /\* if(f->buffersent == 0){ *\/ */
     /* 	   /\*   //memset (buf, '0x55', bytes); *\/ */
     /* 	   /\*   memset (buf, 0, bytes); *\/ */
     /* 	   /\*   i=0; *\/ */
     /* 	   /\*   while(bytes < buflen*0.1){ *\/ */
     /* 	   /\*     j = 418*i; *\/ */
     /* 	   /\*     buf[j] = '\xFF'; *\/ */
     /* 	   /\*     buf[j+1] = '\xFB'; *\/ */
     /* 	   /\*     buf[j+2] = '\x92'; *\/ */
     /* 	   /\*     buf[j+3] = '\x04'; *\/ */
     /* 	   /\*     bytes += 418; *\/ */
     /* 	   /\*     i++; *\/ */
     /* 	   /\*   } *\/ */
     /* 	   /\*   bytes = buflen; *\/ */
     /* 	   //printf("Sending fake data %i bytes \n", bytes); */
     /* 	   //} */
     /* 	 }else{ */
     /* 	   //Reset the timeout mechanism */
     /* 	   f->skip_read = 0; */
     /* 	   //Read data out of the FIFO to sent to the radio */
     /* 	   bytes = fifo_read(f, buf, buflen); */
     /* 	   brk = 1; */
     /* 	   f->buffersent=1; */
     /* 	   printf("Writing data from the FIFO. Now holding %i\n", f->unread); */

     /* 	   /\* //FILE* fp = fopen("/home/dif/programs/Ushare/captured.mp3", "w+"); *\/ */
     /* 	   /\* //fwrite(buf, buflen, 1, fp); *\/ */
     /* 	   /\* //fclose(fp); *\/ */
     /* 	   /\* i = 0; *\/ */
     /* 	   /\* for (j=0; j<buflen; j++){ *\/ */

     /* 	   /\*   //ffe0 = 1111111111100000  https://stackoverflow.com/questions/55664301/how-to-find-headers-of-mp3-audio-clip *\/ */

	     
     /* 	   /\*   if(buf[j] == '\xFF' && (buf[j+1] & '\xE0') == '\xE0'){ *\/ */
     /* 	   /\*     //https://www.binaryhexconverter.com/hex-to-binary-converter *\/ */
     /* 	   /\*     printf("Position: %i Since last block: %i :", j, j-i); *\/ */
     /* 	   /\*     printf("%hhx ", buf[j]); *\/ */
     /* 	   /\*     printf("%hhx ", buf[j+1]); *\/ */
     /* 	   /\*     printf("%hhx ", buf[j+2]); *\/ */
     /* 	   /\*     printf("%hhx \n", buf[j+3]); *\/ */
     /* 	   /\*     i = j; *\/ */
	       
     /* 	   /\*   } *\/ */
     /* 	   /\*   //for (i = 0; i < sizeof(bits); i++) { *\/ */
     /* 	   /\*   //  bits[i] = (buf[j] >> i) & mask; *\/ */
     /* 	   /\*   //  printf("j:%i, i%i, bit: %d\n",j, i, bits[i]); *\/ */
     /* 	   /\*   //}      *\/ */
     /* 	   /\* } *\/ */




	   
	   
	   
     /* 	 }  */
     /*   } */
     
     return bytes;
}

// If the FIFO has insufficient data at the beginning we need to create some blank frames which match the same bitrate
// Easiest is to generate them by re-writing some of the ones sent by the server itself






/* // Access the buffer to read it, pass data out for other uses */
/* int AccessFIFO(fifo_t * f, char *buf, int buflen) */
/* { */
/*      int bytes = 0; */
/*      int brk = 0; */
/*      // Target buffer level, but this may not be achievable on low bitrate */
/*      // streams before the radio times out ~25 seconds after making a request. */
/*      // The radio attempts to fill it's own buffer with an initial series of */
/*      // rapid requests, emptying the FIFO at the same time data arrives */
/*      // The buffer length here is the chunk size requested by the radio of */
/*      // 131072 bytes. */
/*      // 56k-bit/sec stream 131072/1000/56=2.3 seconds */
/*      //Sending partial data (<131072 bytes) just seems to stop the radio working */
/*      int balance = buflen+1; //131072*1.75 = 229376 */
/*      int send;  */
/*      //int balance = FIFO_BUFFER_SIZE*0.25; */
/*      /\* size_t readok; *\/ */


/*      /\* //pre-load the FIFO with junk *\/ */
/*      /\* char ZEROARRAY[200000]; *\/ */
     
/*      /\* FILE* fp = fopen("/home/dif/programs/Ushare/5-seconds-of-silence.mp3", "r"); *\/ */
/*      /\* readok = fread(ZEROARRAY, 1, 200000, fp); *\/ */
/*      /\* fifo_write(f, ZEROARRAY, readok); *\/ */
/*      /\* fclose(fp); *\/ */
/*      /\* printf("Put data %li in the FIFO. Now holding %i\n", readok, f->unread); *\/ */

/*      if(f->buffersent==0){ */
/*        sleep(5); */
/*      } */
     
/*      while(!brk && f->skip_read <= 23) */
/*        { */
/* 	 if(f->unread <= balance){ */
/* 	   //Too little data in the FIFO, give more time for data to arrive */
	   
/* 	   f->skip_read++; */
/*            sleep(3); */
/* 	   //Send dummy data to prevent timeout while the buffer builds up */
/* 	   //memset (buf, 0, buflen); */
/* 	   printf("Sending %i bytes - Buffer: %i. Skips: %i\n",f->unread, f->unread, f->skip_read); */
/* 	   bytes = fifo_read(f, buf, 300); */
/* 	 }else{ */
/* 	   //Reset the timeout mechanism */
/* 	   f->skip_read = 0; */
/* 	   //Read data out of the FIFO to sent to the radio */
/* 	   //bytes = fifo_read(f, buf, buflen); */
/* 	   //memset (buf, 0, buflen); */
/* 	   bytes = fifo_read(f, buf, buflen); */
/* 	   printf("Sending full buffer - Buffer: %i. Skips: %i\n", f->unread, f->skip_read); */
/* 	   //bytes = buflen; */
/* 	   brk = 1; */
/* 	 } */
	
	 
/* 	 //Send dummy data to prevent timeout while the buffer builds up */
/* 	 //memset (buf, 0, buflen); */

/* 	 if(f->unread >= balance){ */
/* 	   //Sufficient data in the FIFO */
/* 	   //Reset the timeout mechanism */
/* 	   f->skip_read = 0; */
/* 	   f->buffersent = 1; // switch away from the partial sends used initially.  */
/* 	   //Read data out of the FIFO to sent to the radio */
/* 	   bytes = fifo_read(f, buf, buflen); */
/* 	   printf("Sending full buffer - Buffer: %i. Skips: %i\n", f->unread, f->skip_read); */
/* 	   brk = 1; */
/* 	 }else if((f->skip_read >= 2) && f->buffersent==0){ */
/* 	   //Keep the connection alive with a little data while waiting for FIFO to fill */
/* 	   f->skip_read = 0; */
/* 	   send = 0.02*f->unread; */
/* 	   //Send a proportion of what we've got, up to the maximum.  */
/* 	   bytes = fifo_read(f, buf, (((buflen)<(send))?(buflen):(send))); */
/* 	   printf("Sending: %i. Buffer: %i. Skips: %i\n", bytes, f->unread, f->skip_read); */
/* 	   brk = 1; */
/* 	 }else{ */
/* 	   //Too little data in the FIFO, but the connection won't time out yet */
/* 	   f->skip_read++; */
/*            sleep(2); */
/* 	   printf("Skipping - Buffer: %i. Skips: %i\n", f->unread, f->skip_read); */
/* 	 } */




	 
/*        } */
/*      return bytes; */
/* } */

// Read from stream into the buffer
void * ReadStream(void * param)
{
     pthread_detach(pthread_self());
  
     fifo_t * f = (fifo_t *)param;     
     ssize_t nn = 0;
     f->tid = pthread_self();
     f->startingup = 0; // Real threadID known, reset start-up flag.
     
     struct sockaddr_in sin;

     memset((char*)&sin,0,sizeof(sin));
     sin.sin_family=AF_INET;
     
     #define READ_STREAM_CHUNK 4096
     
     char temp[READ_STREAM_CHUNK];

     int n=snprintf(temp, READ_STREAM_CHUNK, "%s", f->url);
     if(n==-1 || n >= READ_STREAM_CHUNK)
       temp[sizeof(temp)-1]=0;
     
     char* host=temp;
     
     char* res=strchr(host,'/');
     if(res)
       { *res=0; res++; }
     else
       res=(char*)"";
     
     char* port=strchr(host,':');
     if(port)
       { *port=0; port++; }
     else
       port=(char*)"80";
     
     sin.sin_port=htons(atoi(port));
     sin.sin_addr.s_addr=inet_addr(host);
     
     if(sin.sin_addr.s_addr==INADDR_NONE)
       {
	 struct hostent* he=gethostbyname(host);
	 if(!he)
	   {
	     if(f->verbose)
	       printf("Host is not found: %s\n",host);
	     f->error = -2;
	     f->tid = (pthread_t)NULL;
	     pthread_exit(NULL);
	   }
	 memcpy((char*)&sin.sin_addr.s_addr,he->h_addr,sizeof(sin.sin_addr.s_addr));
       }
     
     int sock=socket(PF_INET,SOCK_STREAM,0);
     
     if(sock==-1)
       {
	 if(f->verbose)
	   printf("Socket() fail\n");
	 f->error = -3;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     
     if(connect(sock, (struct sockaddr*)&sin, sizeof(sin)))
       {
	 if(f->verbose)
	   printf("Can`t connect to %s:%s\n",host,port);
	 close(sock);
	 f->error = -4;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     
     int on=1;
       setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));

     FILE* sfd=fdopen(sock,"a+");
     
     if(!sfd)
       {
	 if(f->verbose)
	   printf("fdopen() fail\n");
	 close(sock);
	 f->error = -5;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     if(f->verbose)
       printf("Connected to %s:%s\n",host,port);

     //Must specify HTTP/1.0 here otherwise some servers will reply with chunked encoding which isn't handled here
     fprintf(sfd,"GET /%s HTTP/1.0\r\nHost: %s:%s\r\nConnection: close\r\nUser-Agent:Reciva radio next generation\r\n\r\n",res,host,port);    
     
     fflush(sfd);
     
     int lineno=0;
     
     int status_code=-1;
     
     while(fgets(temp,sizeof(temp),sfd))
       {
	 char* p=strpbrk(temp,"\r\n");
	 //printf("Header %s\n", temp);  
	 if(p)
	   *p=0;
	 
	 if(!*temp)
	   break;
	 
	 if(!lineno)
	   {
	     p=strchr(temp,' ');
	     if(p)
	       {
		 //printf("Header %s\n", p);   
		 while(*p && *p==' ')
		   p++;
		 char* p2=strchr(p,' ');
		 if(p2)
		   {
		     *p2=0;
		     status_code=atoi(p);
		   }
	       }
	   }
	 
	 lineno++;
       }
     
     if(status_code!=200)
       {
	 if(f->verbose)
	   printf("Status code: %i\n",status_code);
	 
	 fclose(sfd);
	 f->error = -6;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     
     if(f->verbose)
       printf("Streaming http://%s\n",f->url);
     
     int brk=0;
     
     int len;

     ///////////// Implement chunked encoding around here ///////////////////////
     while(!brk && !f->terminate && (len=fread(temp,1,sizeof(temp),sfd))>0)
       {
	 int l=0;
 
	 while(!brk && !f->terminate && l<len)
	   {
	     // If the buffer cannot hold the latest incoming data just skip this data
	     // Don't attempt to pause and catch up - the buffer size is just too small to deal with this
	     // Record how many times in a row data is skipped. More than 30 skips suggests the end use
	     // on the radio has gone away - the whole process can be terminated at that time. 
	     if (f->unread + len-l >= f->size){
	       // Skip this data
	       f->skip_write++;
	       //nn = len-l;
	     }else{
	       // Things are back on track - forget the missed data and restart
	       f->skip_write = 0;
	       nn = fifo_write(f, temp+l, len-l);
	     }

	     if(f->verbose)
	       printf("Balance in buffer: %i\n", f->unread);
	     
	     if(f->skip_write > 20)
	       {
		 if(f->verbose)
		   printf("Radio has gone away!\n");
		 brk=1;
	       }else{
	         l+=nn;
	       }
	   }
       }

     fclose(sfd);
     close(sock);
      
     if(f->verbose)
       printf("Disconnected\n");
     
     f->error = 0;
     f->terminate = 0;
     f->tid = (pthread_t)NULL;
     pthread_exit(NULL);
}

//This initializes the FIFO structure with the given buffer and size
void fifo_init(fifo_t * f, char * url, int upnp_id){

  //Clear out any old data from a previous stream
  memset(f->buf, 0, FIFO_BUFFER_SIZE);
  memset(f->url, 0, FIFO_URL_SIZE);
  
  //If there's any thread ID already stored and still running close that now
  if(f->tid)
    pthread_cancel(f->tid);
  
  f->head = 0;
  f->tail = 0;
  f->size = FIFO_BUFFER_SIZE;
  f->unread = 0;
  f->tid = (pthread_t)NULL;
  f->error = 0;
  strcpy(f->url, url);
  f->upnp_id = upnp_id;
  f->verbose = 0;
  f->skip_write = 0;
  f->skip_read = 0;
  f->initialise = 2;
  f->terminate = 0;
  f->startingup = 1; 
  f->buffersent = 0; 
} 

//This reads nbytes bytes from the FIFO
//The number of bytes read is returned
int fifo_read(fifo_t * f, void * buf, int nbytes){
     int i;
     char * p;
     
     pthread_mutex_lock( &mutex1 );
     
     p = buf;
    
     for(i=0; i < nbytes; i++){
          if( f->tail != f->head ){ //see if any data is available
               *p++ = f->buf[f->tail];  //grab a byte from the buffer
               f->tail++;  //increment the tail
               if( f->tail == f->size ){  //check for wrap-around
                    f->tail = 0;
               }
          } else {
               return i; //number of bytes read
          }
     }
     f->unread -= nbytes;

     pthread_mutex_unlock( &mutex1 );
     
     return nbytes;
}

//This writes up to nbytes bytes to the FIFO
//If the head runs in to the tail, not all bytes are written
//The number of bytes written is returned
int fifo_write(fifo_t * f, const void * buf, int nbytes){
     int i;
     const char * p;

     pthread_mutex_lock( &mutex1 );
     
     p = buf;
     for(i=0; i < nbytes; i++){
           //first check to see if there is space in the buffer
       if( (f->head + 1 == f->tail) || ( (f->head + 1 == f->size) && (f->tail == 0) ) ) 
	 {
	   f->unread += i;
	   return i; //no more room
	 } else {
	   f->buf[f->head] = *p++;
	 
	   f->head++;  //increment the head
	 
	   if( f->head == f->size ){  //check for wrap-around
	     f->head = 0;
	   }
         }
     }
         
     f->unread += nbytes;
     
     pthread_mutex_unlock( &mutex1 );
          
     return nbytes;
}
