/*
 * http.c : GeeXboX uShare Web Server handler.
 * Originally developped for the GeeXboX project.
 * Parts of the code are originated from GMediaServer from Oskar Liljeblad.
 * Copyright (C) 2005-2007 Benjamin Zores <ben@geexbox.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <pthread.h> 
#include <signal.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <upnp/upnp.h>
#include <upnp/upnptools.h>

#include "services.h"
#include "cds.h"
#include "cms.h"
#include "msr.h"
#include "metadata.h"
#include "http.h"
#include "minmax.h"
#include "trace.h"
#include "presentation.h"
#include "osdep.h"
#include "mime.h"
#include "fifo.h"

#define PROTOCOL_TYPE_PRE_SZ  11   /* for the str length of "http-get:*:" */
#define PROTOCOL_TYPE_SUFF_SZ 2    /* for the str length of ":*" */

struct web_file_t {
  char *fullpath;
  off_t pos;
  enum {
    FILE_LOCAL,
    FILE_MEMORY,
    FILE_STREAM
  } type;
  union {
    struct {
      int fd;
      struct upnp_entry_t *entry;
    } local;
    struct {
      char *contents;
      off_t len;
    } memory;
  } detail;
};


static inline void
set_info_file (struct File_Info *info, const size_t length,
               const char *content_type)
{
  info->file_length = length;
  info->last_modified = 0;
  info->is_directory = 0;
  info->is_readable = 1;
  info->content_type = ixmlCloneDOMString (content_type);
}

static int
http_get_info (const char *filename, struct File_Info *info)
{
  extern struct ushare_t *ut;
  struct upnp_entry_t *entry = NULL;
  struct stat st;
  int upnp_id = 0;
  char *content_type = NULL;
  char *protocol = NULL;
  
  if (!filename || !info)
    return -1;

  log_verbose ("http_get_info, filename : %s\n", filename);

  if (!strcmp (filename, CDS_LOCATION))
  {
    set_info_file (info, CDS_DESCRIPTION_LEN, SERVICE_CONTENT_TYPE);
    return 0;
  }

  if (!strcmp (filename, CMS_LOCATION))
  {
    set_info_file (info, CMS_DESCRIPTION_LEN, SERVICE_CONTENT_TYPE);
    return 0;
  }

  if (!strcmp (filename, MSR_LOCATION))
  {
    set_info_file (info, MSR_DESCRIPTION_LEN, SERVICE_CONTENT_TYPE);
    return 0;
  }

  if (ut->use_presentation && !strcmp (filename, USHARE_PRESENTATION_PAGE))
  {
    if (build_presentation_page (ut) < 0)
      return -1;

    set_info_file (info, ut->presentation->len, PRESENTATION_PAGE_CONTENT_TYPE);
    return 0;
  }

  if (ut->use_presentation && !strncmp (filename, USHARE_CGI, strlen (USHARE_CGI)))
  {
    if (process_cgi (ut, (char *) (filename + strlen (USHARE_CGI) + 1)) < 0)
      return -1;

   set_info_file (info, ut->presentation->len, PRESENTATION_PAGE_CONTENT_TYPE);
    return 0;
  }

  upnp_id = atoi (strrchr (filename, '/') + 1);
  entry = upnp_get_entry (ut, upnp_id);
  if (!entry)
    return -1;

  if (!entry->fullpath)
    return -1;

  if (stat (entry->fullpath, &st) < 0)
    return -1;

  if (access (entry->fullpath, R_OK) < 0)
  {
    if (errno != EACCES)
      return -1;
    info->is_readable = 0;
  }
  else
    info->is_readable = 1;

  /* file exist and can be read */
  info->file_length = st.st_size;
  info->last_modified = st.st_mtime;
  info->is_directory = S_ISDIR (st.st_mode);

  //Override the size entry for m3u files - we will supply a stream, not data from the file
  if(!strcmp(strrchr(filename, '\0') - 4, ".m3u")){
    info->file_length = (off_t)FIFO_STREAM_LENGTH;
  }
    
  protocol = 
#ifdef HAVE_DLNA
    entry->dlna_profile ?
    dlna_write_protocol_info (DLNA_PROTOCOL_INFO_TYPE_HTTP,
                              DLNA_ORG_PLAY_SPEED_NORMAL,
                              DLNA_ORG_CONVERSION_NONE,
                              DLNA_ORG_OPERATION_RANGE,
                              ut->dlna_flags, entry->dlna_profile) :
#endif /* HAVE_DLNA */
    mime_get_protocol (entry->mime_type);

  content_type =
    strndup ((protocol + PROTOCOL_TYPE_PRE_SZ),
             strlen (protocol + PROTOCOL_TYPE_PRE_SZ)
             - PROTOCOL_TYPE_SUFF_SZ);
  free (protocol);

  if (content_type)
  {
    info->content_type = ixmlCloneDOMString (content_type);
    free (content_type);
  }
  else
    info->content_type = ixmlCloneDOMString ("");

  return 0;
}

static UpnpWebFileHandle
get_file_memory (const char *fullpath, const char *description,
                 const size_t length)
{
  struct web_file_t *file;

  file = malloc (sizeof (struct web_file_t));
  file->fullpath = strdup (fullpath);
  file->pos = 0;
  file->type = FILE_MEMORY;
  file->detail.memory.contents = strdup (description);
  file->detail.memory.len = length;

  return ((UpnpWebFileHandle) file);
}

static UpnpWebFileHandle
http_open (const char *filename, enum UpnpOpenFileMode mode)
{
  extern struct ushare_t *ut;
  struct upnp_entry_t *entry = NULL;
  struct web_file_t *file;
  int fd, upnp_id = 0;

  if (!filename)
    return NULL;

  log_verbose ("http_open, filename : %s\n", filename);

  if (mode != UPNP_READ)
    return NULL;

  if (!strcmp (filename, CDS_LOCATION))
    return get_file_memory (CDS_LOCATION, CDS_DESCRIPTION, CDS_DESCRIPTION_LEN);

  if (!strcmp (filename, CMS_LOCATION))
    return get_file_memory (CMS_LOCATION, CMS_DESCRIPTION, CMS_DESCRIPTION_LEN);

  if (!strcmp (filename, MSR_LOCATION))
    return get_file_memory (MSR_LOCATION, MSR_DESCRIPTION, MSR_DESCRIPTION_LEN);

  if (ut->use_presentation && ( !strcmp (filename, USHARE_PRESENTATION_PAGE)
      || !strncmp (filename, USHARE_CGI, strlen (USHARE_CGI))))
    return get_file_memory (USHARE_PRESENTATION_PAGE, ut->presentation->buf,
                            ut->presentation->len);

  upnp_id = atoi (strrchr (filename, '/') + 1);
  entry = upnp_get_entry (ut, upnp_id);
  if (!entry)
    return NULL;

  if (!entry->fullpath)
    return NULL;

  log_verbose ("Fullpath : %s\n", entry->fullpath);

  // For streams to be buffered and relayed from the net - not serving m3u files  
  if(!strcmp(strrchr(filename, '\0') - 4, ".m3u")){
    
    int i = 0;
    
    if(my_fifo->tid && my_fifo->upnp_id!=upnp_id){
      //Existing stream download thread is running but upnp_id has changed
      //indicating URL requested has changed - close that existing connection 
      my_fifo->terminate = 1;
      
      while(my_fifo->tid){
	sleep (1);
	if(i > 15){
	  //Graceful termination has failed, try this instead. 
	  pthread_cancel(my_fifo->tid);
	  my_fifo->tid = (pthread_t)NULL;
	}
	i++;
      }
    }
    
    if(my_fifo->tid==(pthread_t)NULL && my_fifo->startingup==0){
      // No streaming download exists - initialise fifo and create thread.
      
      my_fifo->startingup = 1; // Prevent multiple rapid requests before threadID known
      
      // Find the URL to start streaming
      FILE* fp=fopen(entry->fullpath,"r");
      if (!fp){
	my_fifo->startingup = 0;
	return NULL;
      }
      
      char buf[256];
      char url[FIFO_URL_SIZE];
      int line = 0;
      
      url[0] = '\0';
      
      while(fgets(buf,sizeof(buf),fp)){
	line++;
	if(line==entry->line && !strncmp(buf, "http://", 7)){
	  sprintf (url, "%.*s", FIFO_URL_SIZE-1, buf+7);
	  url[strcspn(url, "\r\n")] = '\0'; // Remote new lines & line feeds
	  break;
	}
      }
      
      fclose(fp);
      
      if(strlen(url)==0){
	my_fifo->startingup = 0;
	return NULL;
      }

      // Initialise fifo and create thread. 
      fifo_init(my_fifo, url, upnp_id); 
      pthread_create(&my_fifo->tid, NULL, ReadStream, (void*)my_fifo);
    }
    
    file = malloc (sizeof (struct web_file_t));
    file->fullpath = strdup (entry->fullpath);
    file->pos = 0;
    file->type = FILE_STREAM;
    file->detail.local.entry = entry;
    file->detail.local.fd = (size_t)NULL;
    
  }else{
    // For real files shared from the drive
    fd = open (entry->fullpath, O_RDONLY | O_NONBLOCK | O_SYNC | O_NDELAY);
    if (fd < 0)
      return NULL;
    
    file = malloc (sizeof (struct web_file_t));
    file->fullpath = strdup (entry->fullpath);
    file->pos = 0;
    file->type = FILE_LOCAL;
    file->detail.local.entry = entry;
    file->detail.local.fd = fd; 
  }
  
  return ((UpnpWebFileHandle) file);
}

static int
http_read (UpnpWebFileHandle fh, char *buf, size_t buflen)
{
  struct web_file_t *file = (struct web_file_t *) fh;
  ssize_t len = -1;

  log_verbose ("http_read\n");

  if (!file)
    return -1;

  switch (file->type)
  {
  case FILE_LOCAL:
    log_verbose ("Read local file.\n");
    len = read (file->detail.local.fd, buf, buflen);
    break;
  case FILE_MEMORY:
    log_verbose ("Read file from memory.\n");
    len = (size_t) MIN (buflen, file->detail.memory.len - file->pos);
    memcpy (buf, file->detail.memory.contents + file->pos, (size_t) len);
    break;
  case FILE_STREAM:
    log_verbose ("Reading internet stream(1).\n");
    
    if(my_fifo->initialise == 1 || my_fifo->initialise == 2){
      //Stage 1 initial data - the radio wants to be sent 128k inc ID3 tag, but is happy with null data
      //Stage 2 initial data - radio wants the last 128 bytes of the 'file', but is happy with null data
      memset (buf, 0, buflen);
      len = buflen;
      my_fifo->initialise--;
      //if(my_fifo->initialise == 0)
	sleep(2.5);
    }else{
      log_verbose("Sending data. Remaining buffer: %i.\n", my_fifo->unread);
      len = AccessFIFO(my_fifo, buf, buflen);
    }
    
    log_verbose ("Reading internet stream(2).\n");
    
    break;
  default:
    log_verbose ("Unknown file type.\n");
    break;
  }

  if (len >= 0)
    file->pos += len;

  log_verbose ("Read %zd bytes.\n", len);

  return len;
}

static int
http_write (UpnpWebFileHandle fh __attribute__((unused)),
            char *buf __attribute__((unused)),
            size_t buflen __attribute__((unused)))
{
  log_verbose ("http write\n");

  return 0;
}

static int
http_seek (UpnpWebFileHandle fh, off_t offset, int origin)
{
  struct web_file_t *file = (struct web_file_t *) fh;
  off_t newpos = -1;

  log_verbose ("http_seek\n");
  
  if (!file)
    return -1;

  switch (origin)
  {
  case SEEK_SET:
    log_verbose ("Attempting to seek to %ld (was at %ld) in %s\n",
                offset, file->pos, file->fullpath);
    newpos = offset;
    break;
  case SEEK_CUR:
    log_verbose ("Attempting to seek by %ld from %ld in %s\n",
                offset, file->pos, file->fullpath);
    newpos = file->pos + offset;
    break;
  case SEEK_END:
    log_verbose ("Attempting to seek by %ld from end (was at %ld) in %s\n",
                offset, file->pos, file->fullpath);

    if (file->type == FILE_LOCAL)
    {
      struct stat sb;
      if (stat (file->fullpath, &sb) < 0)
      {
        log_verbose ("%s: cannot stat: %s\n",
                    file->fullpath, strerror (errno));
        return -1;
      }
      newpos = sb.st_size + offset;
    }
    else if (file->type == FILE_MEMORY)
      newpos = file->detail.memory.len + offset;
    else if (file->type == FILE_STREAM)
      newpos = FIFO_STREAM_LENGTH + offset;
    break;
  }

  switch (file->type)
  {
  case FILE_LOCAL:
    /* Just make sure we cannot seek before start of file. */
    if (newpos < 0)
    {
      log_verbose ("%s: cannot seek: %s\n", file->fullpath, strerror (EINVAL));
      return -1;
    }

    /* Don't seek with origin as specified above, as file may have
       changed in size since our last stat. */
    if (lseek (file->detail.local.fd, newpos, SEEK_SET) == -1)
    {
      log_verbose ("%s: cannot seek: %s\n", file->fullpath, strerror (errno));
      return -1;
    }
    break;
  case FILE_MEMORY:
    if (newpos < 0 || newpos > file->detail.memory.len)
    {
      log_verbose ("%s: cannot seek: %s\n", file->fullpath, strerror (EINVAL));
      return -1;
    }
    break;
  case FILE_STREAM:
    //Does nothing for a stream 
    break;
  }

  file->pos = newpos;

  return 0;
}

static int
http_close (UpnpWebFileHandle fh)
{
  struct web_file_t *file = (struct web_file_t *) fh;

  log_verbose ("http_close\n");

  if (!file)
    return -1;

  switch (file->type)
  {
  case FILE_LOCAL:
    close (file->detail.local.fd);
    break;
  case FILE_MEMORY:
    /* no close operation */
    if (file->detail.memory.contents)
      free (file->detail.memory.contents);
    break;
  case FILE_STREAM:
    break;
  default:
    log_verbose ("Unknown file type.\n");
    break;
  }

  if (file->fullpath)
    free (file->fullpath);
  free (file);

  return 0;
}

struct UpnpVirtualDirCallbacks virtual_dir_callbacks =
  {
    http_get_info,
    http_open,
    http_read,
    http_write,
    http_seek,
    http_close
  };

// Access the buffer to read it, pass data out for other uses
int AccessFIFO(fifo_t * f, char *buf, int buflen)
{
     int bytes = 0;
     int balance = buflen*1.05;

     //Catch the case that the FIFO has been destroyed in the other thread
     if(f->unread == 0 || f->tid == (pthread_t)NULL)
       return 0;
     
     if(f->unread <= balance){
       //Too little data in the FIFO, but send something to avoid a timeout
       sleep(3);
       bytes = fifo_read(f, buf, 1);
     }else{
       //Read data out of the FIFO to sent to the radio
       bytes = fifo_read(f, buf, buflen);
     }
     //printf("Sending data: %i. Buffer: %i. Skipped reads: %i\n", bytes, f->unread, f->skip_read);
     
     return bytes;
}

// Read from stream into the buffer
void * ReadStream(void * param)
{
     pthread_detach(pthread_self());
  
     fifo_t * f = (fifo_t *)param;     
     ssize_t nn = 0;
     f->tid = pthread_self();
     f->startingup = 0; // Real threadID known, reset start-up flag.
     
     struct sockaddr_in sin;

     memset((char*)&sin,0,sizeof(sin));
     sin.sin_family=AF_INET;
     
     #define READ_STREAM_CHUNK 4096
     
     char temp[READ_STREAM_CHUNK];

     int n=snprintf(temp, READ_STREAM_CHUNK, "%s", f->url);
     if(n==-1 || n >= READ_STREAM_CHUNK)
       temp[sizeof(temp)-1]=0;
     
     char* host=temp;
     
     char* res=strchr(host,'/');
     if(res)
       { *res=0; res++; }
     else
       res=(char*)"";
     
     char* port=strchr(host,':');
     if(port)
       { *port=0; port++; }
     else
       port=(char*)"80";
     
     sin.sin_port=htons(atoi(port));
     sin.sin_addr.s_addr=inet_addr(host);
     
     if(sin.sin_addr.s_addr==INADDR_NONE)
       {
	 struct hostent* he=gethostbyname(host);
	 if(!he)
	   {
	     if(f->verbose)
	       printf("Host is not found: %s\n",host);
	     f->error = -2;
	     f->tid = (pthread_t)NULL;
	     pthread_exit(NULL);
	   }
	 memcpy((char*)&sin.sin_addr.s_addr,he->h_addr,sizeof(sin.sin_addr.s_addr));
       }
     
     int sock=socket(PF_INET,SOCK_STREAM,0);
     
     if(sock==-1)
       {
	 if(f->verbose)
	   printf("Socket() fail\n");
	 f->error = -3;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     
     if(connect(sock, (struct sockaddr*)&sin, sizeof(sin)))
       {
	 if(f->verbose)
	   printf("Can`t connect to %s:%s\n",host,port);
	 close(sock);
	 f->error = -4;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     
     int on=1;
       setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));

     FILE* sfd=fdopen(sock,"a+");
     
     if(!sfd)
       {
	 if(f->verbose)
	   printf("fdopen() fail\n");
	 close(sock);
	 f->error = -5;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     if(f->verbose)
       printf("Connected to %s:%s\n",host,port);

     //Must specify HTTP/1.0 here otherwise some servers will reply with chunked encoding which isn't handled here
     fprintf(sfd,"GET /%s HTTP/1.0\r\nHost: %s:%s\r\nConnection: close\r\nUser-Agent:Reciva radio next generation\r\n\r\n",res,host,port);    
     
     fflush(sfd);
     
     int lineno=0;
     
     int status_code=-1;
     
     while(fgets(temp,sizeof(temp),sfd))
       {
	 char* p=strpbrk(temp,"\r\n");
	 //printf("Header %s\n", temp);  
	 if(p)
	   *p=0;
	 
	 if(!*temp)
	   break;
	 
	 if(!lineno)
	   {
	     p=strchr(temp,' ');
	     if(p)
	       {
		 //printf("Header %s\n", p);   
		 while(*p && *p==' ')
		   p++;
		 char* p2=strchr(p,' ');
		 if(p2)
		   {
		     *p2=0;
		     status_code=atoi(p);
		   }
	       }
	   }
	 
	 lineno++;
       }
     
     if(status_code!=200)
       {
	 if(f->verbose)
	   printf("Status code: %i\n",status_code);
	 
	 fclose(sfd);
	 f->error = -6;
	 f->tid = (pthread_t)NULL;
	 pthread_exit(NULL);
       }
     
     if(f->verbose)
       printf("Streaming http://%s\n",f->url);
     
     int brk=0;
     
     int len;

     ///////////// Implement chunked encoding around here ///////////////////////
     while(!brk && !f->terminate && (len=fread(temp,1,sizeof(temp),sfd))>0)
       {
	 int l=0;
 
	 while(!brk && !f->terminate && l<len)
	   {
	     // If the buffer cannot hold the latest incoming data just skip this data
	     // Don't attempt to pause and catch up - the buffer size is just too small to deal with this
	     // Record how many times in a row data is skipped. More than 30 skips suggests the end use
	     // on the radio has gone away - the whole process can be terminated at that time. 
	     if (f->unread + len-l >= f->size){
	       // Skip this data
	       f->skip_write++;
	       //nn = len-l;
	     }else{
	       // Things are back on track - forget the missed data and restart
	       f->skip_write = 0;
	       nn = fifo_write(f, temp+l, len-l);
	     }

	     if(f->verbose)
	       printf("Balance in buffer: %i\n", f->unread);
	     
	     if(f->skip_write > 20)
	       {
		 if(f->verbose)
		   printf("Radio has gone away!\n");
		 brk=1;
	       }else{
	         l+=nn;
	       }
	   }
       }

     fclose(sfd);
     close(sock);
      
     if(f->verbose)
       printf("Disconnected\n");
     
     f->error = 0;
     f->terminate = 0;
     f->startingup = 0;
     f->tid = (pthread_t)NULL;
     pthread_exit(NULL);
}

//This initializes the FIFO structure with the given buffer and size
void fifo_init(fifo_t * f, char * url, int upnp_id){
  
  //If there's any thread ID already stored and still running close that now
  if(f->tid){
    pthread_cancel(f->tid);
  }
  
  //Clear out any old data from a previous stream
  memset(f->buf, 0, FIFO_BUFFER_SIZE);
  memset(f->url, 0, FIFO_URL_SIZE);
  
  f->head = 0;
  f->tail = 0;
  f->size = FIFO_BUFFER_SIZE;
  f->unread = 0;
  f->tid = (pthread_t)NULL;
  f->error = 0;
  strcpy(f->url, url);
  f->upnp_id = upnp_id;
  f->verbose = 0;
  f->skip_write = 0;
  f->skip_read = 0;
  f->initialise = 2;
  f->terminate = 0;
  f->startingup = 1;
} 

//This reads nbytes bytes from the FIFO
//The number of bytes read is returned
int fifo_read(fifo_t * f, void * buf, int nbytes){
     int i;
     char * p;
     
     pthread_mutex_lock( &mutex1 );
     
     p = buf;
    
     for(i=0; i < nbytes; i++){
          if( f->tail != f->head ){ //see if any data is available
               *p++ = f->buf[f->tail];  //grab a byte from the buffer
               f->tail++;  //increment the tail
               if( f->tail == f->size ){  //check for wrap-around
                    f->tail = 0;
               }
          } else {
	       f->unread -= i;
	       pthread_mutex_unlock( &mutex1 );
               return i; //number of bytes read
          }
     }
     
     f->unread -= nbytes;
     pthread_mutex_unlock( &mutex1 );
     return nbytes;
}

//This writes up to nbytes bytes to the FIFO
//If the head runs in to the tail, not all bytes are written
//The number of bytes written is returned
int fifo_write(fifo_t * f, const void * buf, int nbytes){
     int i;
     const char * p;

     pthread_mutex_lock( &mutex1 );
     
     p = buf;
     for(i=0; i < nbytes; i++){
           //first check to see if there is space in the buffer
       if( (f->head + 1 == f->tail) || ( (f->head + 1 == f->size) && (f->tail == 0) ) ) 
	 {
	   f->unread += i;
	   pthread_mutex_unlock( &mutex1 );
	   return i; //no more room
	 } else {
	   f->buf[f->head] = *p++;
	 
	   f->head++;  //increment the head
	 
	   if( f->head == f->size ){  //check for wrap-around
	     f->head = 0;
	   }
         }
     }
         
     f->unread += nbytes;
     pthread_mutex_unlock( &mutex1 );
     return nbytes;
}

